package com.devbeh;


import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;

import com.devbeh.commondata.AtomicKnowledge;
import com.devbeh.commondata.DevBehDetections;
import com.devbeh.commondata.DevBehEntry;
import com.devbeh.commondata.DevBehFrameObject;
import com.devbeh.commondata.Empty;
import com.devbeh.commondata.KBAction;
import com.devbeh.commondata.KBBox;
import com.devbeh.commondata.KBEntry;
import com.devbeh.commondata.KBObject;
import com.devbeh.commondata.KBPoint2d;
import com.devbeh.commondata.KBPoint3d;
import com.devbeh.commondata.KBServiceCommand;
import com.devbeh.commondata.KBServiceResponse;
import org.jpl7.Atom;
import org.jpl7.Compound;
import org.jpl7.Query;
import org.jpl7.Term;


public class KBWorkerStarter
{
    /**
     * Используется для чтения команд.
     */
    private static final InputStream KB_IN = new BufferedInputStream(System.in);

    /**
     * Используется для записи результатов выполнения команд.
     */
    private static final OutputStream KB_OUT = new BufferedOutputStream(System.out);

    // Шаблоны создания термов
    private static final Term OBJ_INFO_TERM = Term.textToTerm("timeBetween(?, ?, Time), correctedBounds(?, box(TLx, TLy, BRx, BRy), Time)");
    private static final Term BOUNDS_TERM = Term.textToTerm("bounds(?, box(?, ?, ?, ?), ?)");
    private static final Term POSITION_TERM = Term.textToTerm("position(?, point(?, ?, ?), ?)");
    private static final Term PERSON_TERM = Term.textToTerm("person(?, ?)");
    private static final Term VEHICLE_TERM = Term.textToTerm("vehicle(?, ?, ?)");
    private static final Term ANIMAL_TERM = Term.textToTerm("animal(?, ?, ?)");
    private static final Term BAG_TERM = Term.textToTerm("bag(?, ?)");
    private static final Term PHONE_TERM = Term.textToTerm("phone(?, ?)");
    private static final Term CHILD_TERM = Term.textToTerm("child(?, ?)");
    private static final Term LOOK_DIRECTION_TERM = Term.textToTerm("lookDirection(?, vector(?, ?, ?), ?)");
    // [point(x1, y1, z1), point(x2, y2, z2)]
    private static final Term VOLUMETRIC_POSITION_TERM = Term.textToTerm("volumetricPosition(?, ?, ?)");

    /**
     * Содержит экземпляры детектированного девиантного поведения. Используется для предотвращения повторного учета
     * одного экземпляра девиантного поведения. В качестве ключа записан терм девиантного поведения, в качестве
     * значения - идентификатор кадра завершения детектированной сцены.
     */
    private static final Map<Term, Long> DEV_BEH_DETECTIONS = new LinkedHashMap<>();

    /**
     * Содержит динамические знания, записанные в БЗ. Очередь используется для удаления из бз устаревших знаний.
     */
    private static Queue<FrameKnowledge> KNOWLEDGE_WINDOW;

    /**
     * Размер окна времени, за которое знания хранятся в бз.
     */
    private static int KB_HISTORY_FRAMES;

    /**
     * Количество кадров, в течение которого повторное детектирование девиантного поведения будет проигнорировано.
     */
    private static int DEV_BEH_IGNORE_PERIOD_FRAMES;

    /**
     * Приблизительный интервал между кадрами в миллисекундах.
     */
    private static int FRAME_INTERVAL_MS;

    /**
     * Идентификатор последнего обработанного кадра.
     */
    private static long LAST_FRAME_ID = 0;

    /**
     * Используется для
     */
    private static PrintWriter ATOMIC_KNOWLEDGE_OUT = null;


    /**
     * Принимает:
     * - путь к файлу пролога, содержащему общие статические знания;
     * - путь к файлу пролога, содержащему специфичные для данного видеопотока статические знания или пустая строка;
     * - размер окна истории для логического вывода в кадрах;
     * - период в кадрах, в течение которого будут игнорироваться аналогичные детектируемые инстансы девиантного
     * поведения;
     * - fps видеопотока в виде double числа.
     */
    public static void main(String[] args)
    {
        try
        {
            run(args);
        }
        catch (Exception e)
        {
            // Передаем сообщение об ошибке родительскому процессу
            e.printStackTrace();
        }
    }


    /**
     * Инициализирует БЗ и обрабатывает запросы от родительского процесса. В случае критической ошибки выбрасывает
     * исключение, завершая тем самым kb-процесс.
     *
     * @param args параметры kb-процесса
     * @throws RuntimeException исключение в случае критической ошибки
     */
    private static void run(String[] args) throws Exception
    {
        if (args.length < 8)
        {
            throw new RuntimeException("Invalid arguments: " + Arrays.toString(args));
        }

        // Считаем, что все проверки параметров были выполнены в родительском процессе
        String generalKbRulesFilePath = args[0];
        String cameraKbRulesFilePath = args[1];
        KB_HISTORY_FRAMES = Integer.parseInt(args[2]);
        KNOWLEDGE_WINDOW = new ArrayBlockingQueue<>(KB_HISTORY_FRAMES + 1);
        DEV_BEH_IGNORE_PERIOD_FRAMES = Integer.parseInt(args[3]);
        int frameHeight = Integer.parseInt(args[4]);
        int frameWidth = Integer.parseInt(args[5]);
        double fps = Double.parseDouble(args[6]);
        FRAME_INTERVAL_MS = (int) Math.round(1000 / fps);

        if (!args[7].isEmpty())
        {
            File atomicKnowledgeFile = new File(args[7]);
            if (!atomicKnowledgeFile.getParentFile().exists())
            {
                if (!atomicKnowledgeFile.getParentFile().mkdirs())
                {
                    throw new RuntimeException("Can't create directories for atomic knowledge file: " +
                            atomicKnowledgeFile.getAbsolutePath());
                }
            }
            ATOMIC_KNOWLEDGE_OUT = new PrintWriter(new BufferedWriter(new OutputStreamWriter(
                    new FileOutputStream(atomicKnowledgeFile))));
        }

        runQuery(new Query("consult", new Term[] { new Atom(generalKbRulesFilePath) }),
                "KB general rules consulting error");
        if (!cameraKbRulesFilePath.isEmpty())
        {
            runQuery(new Query("consult", new Term[] { new Atom(cameraKbRulesFilePath) }),
                    "KB camera specific rules consulting error");
        }

        addAtomicKnowledge(Term.textParamsToTerm("timeStep(?)",
                new Term[] {new org.jpl7.Integer(FRAME_INTERVAL_MS)}));

        addAtomicKnowledge(Term.textParamsToTerm("frameHeight(?)", new Term[] {new org.jpl7.Integer(frameHeight)}));
        addAtomicKnowledge(Term.textParamsToTerm("frameWidth(?)", new Term[] {new org.jpl7.Integer(frameWidth)}));

        // Записываем знания для исправления ошибки отсутствия процедуры
        addAtomicKnowledge(Term.textToTerm("person(p, -1)"));
        addAtomicKnowledge(Term.textToTerm("bounds(p, box(0, 0, 0, 0), -1)"));
        addAtomicKnowledge(Term.textToTerm("position(p, point(0, 0, 0), -1)"));
        addAtomicKnowledge(Term.textToTerm("vehicle(v, v, -1)"));
        addAtomicKnowledge(Term.textToTerm("animal(a, a, -1)"));
        addAtomicKnowledge(Term.textToTerm("bag(b, -1)"));
        addAtomicKnowledge(Term.textToTerm("phone(ph, -1)"));
        addAtomicKnowledge(Term.textToTerm("child(c, -1)"));
        addAtomicKnowledge(Term.textToTerm("lookDirection(p, vector(0, 0, 0), -1)"));
        addAtomicKnowledge(Term.textToTerm("volumetricPosition(v, [], -1)"));

        flushAtomicKnowledge();

        // Родительский процесс ожидает сообщение после успешной инициализации
        writeResponse(KBServiceResponse.newBuilder().setEmpty(Empty.newBuilder().setEmpty(true)).build());

        // Завершение работы происходит посредством убийства процесса
        //noinspection InfiniteLoopStatement
        while (true)
        {
            KBServiceCommand command = readCommand();
            if (command.hasAddFrameKnowledge())
            {
                addFrameKnowledge(command.getAddFrameKnowledge());
            }
            else if (command.hasRunInference())
            {
                runInference();
            }
        }
    }


    /**
     * Записывает атомарное знание в БЗ и файл.
     *
     * @param term терм, записываемый в БЗ и файл
     */
    private static void addAtomicKnowledge(Term term)
    {
        runQuery(new Query(new Compound("asserta", new Term[] {term})),
                "Static knowledge insertion error (" + term.toString() + ")");
        if (ATOMIC_KNOWLEDGE_OUT != null)
        {
            ATOMIC_KNOWLEDGE_OUT.print(term);
            ATOMIC_KNOWLEDGE_OUT.println(".");
        }
    }


    /**
     * Отправляет буферизированные файлы в файл.
     */
    private static void flushAtomicKnowledge()
    {
        if (ATOMIC_KNOWLEDGE_OUT != null)
        {
            ATOMIC_KNOWLEDGE_OUT.flush();
        }
    }


    /**
     * Выполняет переданный запрос, вызывая {@link Query#hasSolution()}. Если метод возвращает false, выбрасывает
     * исключение.
     */
    private static void runQuery(Query q, String errorMsg)
    {
        if (!q.hasSolution())
        {
            throw new RuntimeException(errorMsg);
        }
    }


    /**
     * Считывает команду, отправленную родительским процессом.
     *
     * @return команда
     */
    private static KBServiceCommand readCommand()
    {
        try
        {
            KBServiceCommand command = KBServiceCommand.parseDelimitedFrom(KB_IN);
            if (command == null)
            {
                // Кажется, может произойти только если родительский процесс умрет раньше дочернего
                throw new RuntimeException("Command is null");
            }
            return command;
        }
        catch (IOException e)
        {
            throw new RuntimeException("Exception occurred while reading command", e);
        }
    }


    /**
     * Записывает ответ родительскому процессу.
     *
     * @param response ответ
     */
    private static void writeResponse(KBServiceResponse response)
    {
        try
        {
            response.writeDelimitedTo(KB_OUT);
            KB_OUT.flush();
        }
        catch (IOException e)
        {
            // Кажется, может произойти только если родительский процесс умрет раньше дочернего
            throw new RuntimeException("Exception occurred during writing response");
        }
    }


    /**
     * Добавляет новые знания в бз и удаляет старые.
     *
     * @param knowledge новые знания, относящиеся к очередному кадру
     * @throws RuntimeException в случае возникновения критической ошибки
     */
    private static void addFrameKnowledge(KBEntry knowledge) throws RuntimeException
    {
        List<Term> frameKnowledge = new ArrayList<>();

        long time = knowledge.getFrameId() * FRAME_INTERVAL_MS;

        // Добавляем общие знания кадра
        frameKnowledge.add(new Compound("timeSlice", new Term[] {new org.jpl7.Integer(time)}));

        // Добавляем детектированные знания
        for (AtomicKnowledge ak : knowledge.getKnowledgeList())
        {
            switch (ak.getAkCase())
            {
                case KBOBJECT: addKBObjectKnowledge(ak.getKbObject(), time, frameKnowledge); break;
                case KBACTION: addKBActionKnowledge(ak.getKbAction(), time, frameKnowledge); break;
            }
        }

        // Записываем знания в бз
        for (Term t : frameKnowledge)
        {
            addAtomicKnowledge(t);
        }
        flushAtomicKnowledge();

        LAST_FRAME_ID = knowledge.getFrameId();
        KNOWLEDGE_WINDOW.add(new FrameKnowledge(knowledge.getFrameId(), frameKnowledge));

        // Удаляем устаревшие знания
        while (Objects.requireNonNull(KNOWLEDGE_WINDOW.peek()).frameId + KB_HISTORY_FRAMES <= knowledge.getFrameId())
        {
            FrameKnowledge outdatedKnowledge = KNOWLEDGE_WINDOW.poll();
            for (Term t : outdatedKnowledge.knowledge)
            {
                runQuery(new Query("retract", new Term[] {t}), "KB knowledge retract error");
            }
        }

        // Отправляем сообщение об успешном завершении
        writeResponse(KBServiceResponse.newBuilder().setEmpty(Empty.newBuilder().setEmpty(true)).build());
    }


    /**
     * Преобразует переданные proto объект в термы БЗ.
     *
     * @param kbObject знания в виде proto объекта
     * @param time время кадра
     * @param frameKnowledge список термов для записи
     */
    private static void addKBObjectKnowledge(KBObject kbObject, long time, List<Term> frameKnowledge)
    {
        switch (DetectionClass.findById(kbObject.getClassId()))
        {
            case PERSON: addPersonKnowledge(kbObject, time, frameKnowledge); break;
            case BICYCLE: addVehicleKnowledge(kbObject, "bicycle", time, frameKnowledge); break;
            case MOTORCYCLE: addVehicleKnowledge(kbObject, "motorbike", time, frameKnowledge); break;
            case BUS:
            case TRUCK:
            case CAR: addVehicleKnowledge(kbObject, "car", time, frameKnowledge); break;
            case CAT: addAnimalKnowledge(kbObject, "cat", time, frameKnowledge); break;
            case DOG: addAnimalKnowledge(kbObject, "dog", time, frameKnowledge); break;
            case BACKPACK:
            case SUITCASE:
            case HANDBAG: addBagKnowledge(kbObject, time, frameKnowledge); break;
            case CELL_PHONE: addPhoneKnowledge(kbObject, time, frameKnowledge); break;
        }
    }


    /**
     * Добавляет знания о человеке.
     */
    private static void addPersonKnowledge(KBObject kbObject, long time, List<Term> frameKnowledge)
    {
        frameKnowledge.add(PERSON_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                new org.jpl7.Integer(time)
        }));

        addBoxAndPositionKnowledge(kbObject.getGlobalId(), kbObject.getBoundingBox(), kbObject.getPosition(), time,
                frameKnowledge);

        frameKnowledge.add(LOOK_DIRECTION_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                new org.jpl7.Float(kbObject.getLookDirection().getX()),
                new org.jpl7.Float(kbObject.getLookDirection().getY()),
                new org.jpl7.Float(kbObject.getLookDirection().getZ()),
                new org.jpl7.Integer(time)
        }));

        if (kbObject.getChild())
        {
            frameKnowledge.add(CHILD_TERM.putParams(new Term[] {
                    new org.jpl7.Integer(kbObject.getGlobalId()),
                    new org.jpl7.Integer(time)
            }));
        }
    }


    /**
     * Добавляет знания о средстве передвижения.
     */
    private static void addVehicleKnowledge(KBObject kbObject, String type, long time, List<Term> frameKnowledge)
    {
        frameKnowledge.add(VEHICLE_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                new Atom(type),
                new org.jpl7.Integer(time)
        }));

        addBoxAndPositionKnowledge(kbObject.getGlobalId(), kbObject.getBoundingBox(), kbObject.getPosition(), time,
                frameKnowledge);

        frameKnowledge.add(VOLUMETRIC_POSITION_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                Term.termArrayToList(kbObject.getVolumetricPositionList().stream()
                        .map(p -> new Compound("point", new Term[] {
                                new org.jpl7.Float(p.getX()),
                                new org.jpl7.Float(p.getY()),
                                new org.jpl7.Float(p.getZ())})).toArray(Term[]::new)),
                new org.jpl7.Integer(time)
        }));
    }


    /**
     * Добавляет знания о животном.
     */
    private static void addAnimalKnowledge(KBObject kbObject, String type, long time, List<Term> frameKnowledge)
    {
        frameKnowledge.add(ANIMAL_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                new Atom(type),
                new org.jpl7.Integer(time)
        }));

        addBoxAndPositionKnowledge(kbObject.getGlobalId(), kbObject.getBoundingBox(), kbObject.getPosition(), time,
                frameKnowledge);
    }


    /**
     * Добавляет знания о сумке.
     */
    private static void addBagKnowledge(KBObject kbObject, long time, List<Term> frameKnowledge)
    {
        frameKnowledge.add(BAG_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                new org.jpl7.Integer(time)
        }));

        addBoxAndPositionKnowledge(kbObject.getGlobalId(), kbObject.getBoundingBox(), kbObject.getPosition(), time,
                frameKnowledge);
    }


    private static void addPhoneKnowledge(KBObject kbObject, long time, List<Term> frameKnowledge)
    {
        frameKnowledge.add(PHONE_TERM.putParams(new Term[] {
                new org.jpl7.Integer(kbObject.getGlobalId()),
                new org.jpl7.Integer(time)
        }));

        addBoxAndPositionKnowledge(kbObject.getGlobalId(), kbObject.getBoundingBox(), kbObject.getPosition(), time,
                frameKnowledge);
    }


    /**
     * Добавляет знания об обрамляющей рамке и положении объекта.
     */
    private static void addBoxAndPositionKnowledge(long objId, KBBox boundingBox, KBPoint3d position, long time,
            List<Term> frameKnowledge)
    {
        frameKnowledge.add(BOUNDS_TERM.putParams(new Term[] {
                new org.jpl7.Integer(objId),
                new org.jpl7.Float(boundingBox.getTopLeft().getX()),
                new org.jpl7.Float(boundingBox.getTopLeft().getY()),
                new org.jpl7.Float(boundingBox.getBottomRight().getX()),
                new org.jpl7.Float(boundingBox.getBottomRight().getY()),
                new org.jpl7.Integer(time)}));

        frameKnowledge.add(POSITION_TERM.putParams(new Term[] {
                new org.jpl7.Integer(objId),
                new org.jpl7.Float(position.getX()),
                new org.jpl7.Float(position.getY()),
                new org.jpl7.Float(position.getZ()),
                new org.jpl7.Integer(time)}));
    }


    /**
     * Преобразует переданные proto объект в термы БЗ.
     *
     * @param kbAction знания в виде proto объекта
     * @param time время кадра
     * @param frameKnowledge список термов для записи
     */
    private static void addKBActionKnowledge(KBAction kbAction, long time, List<Term> frameKnowledge)
    {
        boolean started = kbAction.getStarted();
        long p1Id = kbAction.getFirstPersonId();
        long p2Id = kbAction.getSecondPersonId();
        time = time - (long) kbAction.getLatencyFrames() * FRAME_INTERVAL_MS;
        // Пока отсутствуют tugBetween(Robber, Victim) и bendingForward(Person)
        // При удалении устаревших знаний может остаться finished без started. Правила в бз должны корректно
        // обрабатывать такие случаи
        switch (ActionClass.findById(kbAction.getAction()))
        {
            case EXCHANGE_THINGS: addActionKnowledge(started, "transferObject", p1Id, p2Id, time, frameKnowledge); break;
            case TOUCH_POCKET: addActionKnowledge(started, "searchedPockets", p1Id, p2Id, time, frameKnowledge); break;
            case SQUAT_DOWN: addActionKnowledge(started, "squatting", p1Id, p2Id, time, frameKnowledge); break;
            case FALLING_DOWN: addActionKnowledge(started, "falling", p1Id, p2Id, time, frameKnowledge); break;
            case THROW: addActionKnowledge(started, "throwing", p1Id, p2Id, time, frameKnowledge); break;
            case ARM_SWIGS: addActionKnowledge(started, "hittingWithHand", p1Id, p2Id, time, frameKnowledge); break;
            case KICKING_SOMETHING: addActionKnowledge(started, "kicking", p1Id, p2Id, time, frameKnowledge); break;
        }
    }


    /**
     * Добавляет знания о действии.
     */
    private static void addActionKnowledge(boolean started, String actionName, long p1Id, long p2Id, long time,
            List<Term> frameKnowledge)
    {
        frameKnowledge.add(new Compound(started ? "started" : "finished", new Term[] {
                new Compound(actionName, p2Id == 0 ? new Term[]{ new org.jpl7.Integer(p1Id) } :
                        new Term[]{ new org.jpl7.Integer(p1Id), new org.jpl7.Integer(p2Id) }),
                new org.jpl7.Integer(time)}));
    }


    /**
     * Выполняет логический вывод для детектирования девиантного поведения и отправляет результат родительскому
     * процессу.
     *
     * @throws RuntimeException в случае критической ошибки
     */
    private static void runInference() throws RuntimeException
    {
        // Удаляем старые детектированные экземпляры девиантного поведения
        Iterator<Map.Entry<Term, Long>> devBehDetectionsIter = DEV_BEH_DETECTIONS.entrySet().iterator();
        while (devBehDetectionsIter.hasNext())
        {
            Map.Entry<Term, Long> e = devBehDetectionsIter.next();
            if (e.getValue() + DEV_BEH_IGNORE_PERIOD_FRAMES < LAST_FRAME_ID)
            {
                break;
            }
            devBehDetectionsIter.remove();
        }

        // Детектируем девиантное поведение

        KBServiceResponse.Builder response = KBServiceResponse.newBuilder();
        DevBehDetections.Builder detectionBuilder = response.getDetectionsBuilder()
                .setFrameId(LAST_FRAME_ID);
        for (Map<String, Term> devBehParams : new Query(Term.textToTerm("devBeh(DevBeh, StartTime, FinishTime)")))
        {
            Term devBeh = devBehParams.get("DevBeh");
            if (DEV_BEH_DETECTIONS.containsKey(devBeh))
            {
                continue;
            }

            long startTime = devBehParams.get("StartTime").longValue();
            long finishTime = devBehParams.get("FinishTime").longValue();

            long startFrameId = startTime / FRAME_INTERVAL_MS;
            long finishFrameId = finishTime / FRAME_INTERVAL_MS;

            DEV_BEH_DETECTIONS.put(devBeh, finishFrameId);

            DevBehEntry.Builder devBehEntryBuilder = detectionBuilder.addEntriesBuilder()
                    .setType(devBeh.name())
                    .setStartFrameId(startFrameId)
                    .setEndFrameId(finishFrameId)
                    .setInferenceInfo(""); // Пока оставляем данное поле пустым

            // Внутренняя Map на случай повторения детектирования одного объекта в кадре
            Map<Long, Map<Long, DevBehFrameObject>> frameInfoMap = new HashMap<>();
            for (Term obj : devBeh.args())
            {
                for (Map<String, Term> devBehObjInfo : new Query(OBJ_INFO_TERM.putParams(
                                new Term[] { new org.jpl7.Integer(startTime), new org.jpl7.Integer(finishTime), obj })))
                {
                    long frameId = devBehObjInfo.get("Time").longValue() / FRAME_INTERVAL_MS;
                    float topLeftX = devBehObjInfo.get("TLx").floatValue();
                    float topLeftY = devBehObjInfo.get("TLy").floatValue();
                    float bottomRightX = devBehObjInfo.get("BRx").floatValue();
                    float bottomRightY = devBehObjInfo.get("BRy").floatValue();

                    frameInfoMap.computeIfAbsent(frameId, f -> new HashMap<>()).put(
                            obj.longValue(),
                            DevBehFrameObject.newBuilder()
                                    .setId(obj.longValue())
                                    .setBox(KBBox.newBuilder()
                                            .setTopLeft(KBPoint2d.newBuilder()
                                                    .setX(topLeftX)
                                                    .setY(topLeftY))
                                            .setBottomRight(KBPoint2d.newBuilder()
                                                    .setX(bottomRightX)
                                                    .setY(bottomRightY))
                                    ).build());
                }
            }

            frameInfoMap.entrySet().stream().sorted(Map.Entry.comparingByKey())
                    .forEach(frameInfo -> devBehEntryBuilder.addFrameInfoBuilder().setFrameId(frameInfo.getKey())
                            .addAllObjects(frameInfo.getValue().values()));
        }

        // Отправляем результат детектирования
        writeResponse(response.build());
    }


    private static class FrameKnowledge
    {
        /**
         * Идентификатор кадра, к которому относятся знания.
         */
        final long frameId;

        /**
         * Знания, относящиеся к данному кадру.
         */
        final List<Term> knowledge;


        public FrameKnowledge(long frameId, List<Term> knowledge)
        {
            this.frameId = frameId;
            this.knowledge = knowledge;
        }
    }
}
